r"""
Teichmueller curves of Origamis.
"""
# ****************************************************************************
#       Copyright (C) 2011-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from sage.structure.sage_object import SageObject
from sage.structure.element import Element
from sage.structure.parent import Parent

from sage.rings.infinity import Infinity
from sage.rings.integer import Integer
from .origami import Origami

from sage.modular.arithgroup.congroup_sl2z import SL2Z
from sage.modular.arithgroup.arithgroup_perm import Lm,Rm,S2m,S3m,EvenArithmeticSubgroup_Permutation, OddArithmeticSubgroup_Permutation
from copy import copy

class TeichmuellerCurve(SageObject):
    pass

class Cusp(SageObject):
    r"""
    A cusp in a Teichmueller curve.

    - width of the cusp

    - cylinder decomposition

    - lengths

    - heights

    - a representative
    """
    def __init__(self, parent, origami, slope):
        self._parent = parent
        self._origami = origami
        self._slope = slope

        self._width = 1
        L,_ = origami.gl2z_edges()
        oo = L[origami]
        while oo != origami:
            oo = L[oo]
            self._width += 1

        c,l,h = origami.cylinder_diagram(data=True)
        self._cyl_diag = c
        self._lengths = l
        self._heights = h

    def _repr_(self):
        r"""
        String representation
        """
        return "Cusp of %s" %str(parent)

    def width(self):
        r"""
        Return the width of the cusp
        """
        return self._width

    def cylinder_diagram(self,data=False):
        r"""
        Return the cylinder diagram associated to this cusp.
        """
        if data:
            return self._cyl_diag, self._lengths, self._heights
        return self._cyl_diags

    def slope(self):
        r"""
        Return one slope
        """
        return self._slope

def TeichmuellerCurveOfOrigami(origami):
    r"""
    Return the teichmueller curve for an origami

    TESTS::

        sage: from surface_dynamics import *

        sage: o = Origami('(1,2)','(1,3)')
        sage: o.teichmueller_curve() #indirect test
        Teichmueller curve of the origami
        (1)(2,3)
        (1,2)(3)
    """
    origami = origami.to_standard_form()

    dL,dR,dS = origami.sl2z_edges()
    n = len(dL)

    # find numerotation for the elements of the SL2Z-orbit
    mapping = set(dL)
    mapping.remove(origami)
    mapping = sorted(mapping)
    mapping.insert(0,origami)
    inv_mapping = dict((mapping[i],i) for i in range(len(mapping)))

    # consider dL and dR as lists and
    # compute the action of s2 and s3
    # from the one of l and r
    l_edges = [None]*n
    r_edges = [None]*n
    s2_edges = [None]*n
    s3_edges = [None]*n
    for o in mapping:
        i = inv_mapping[o]
        l_edges[i] = inv_mapping[dL[o]]
        r_edges[i] = inv_mapping[dR[o]]
        s2_edges[i] = inv_mapping[dS[o]]
        s3_edges[r_edges[i]] = l_edges[i]

    return TeichmuellerCurveOfOrigami_class(
                mapping, inv_mapping, l_edges, r_edges, s2_edges, s3_edges)

def TeichmuellerCurvesOfOrigamis(origamis, assume_normal_form=False, limit=0, verbose_level=0):
    r"""
    Return a set of Teichmueller curve from a set of origamis

    INPUT:

    - ``origamis`` - iterable of origamis

    - ``assume_normal_form`` - whether or not assume that each origami is in
      normal form

    - ``limit`` - integer (default: 0) - if positive, then stop the computation
      if the size of an orbit is bigger than ``limit``.
    """
    from surface_dynamics.flat_surfaces.origamis.origami_dense import sl2z_orbits

    tcurves = []

    if not assume_normal_form:
        origamis = set(o.relabel() for o in origamis)
    try:
        n = next(iter(origamis)).nb_squares()
    except StopIteration:
        return []

    for dL,dR,dS in sl2z_orbits(origamis,n,limit):
        n = len(dL)

        # find numerotation for the elements of the SL2Z-orbit
        mapping = set(dL)
        mapping = sorted(mapping)
        inv_mapping = dict((mapping[i],i) for i in range(len(mapping)))

        # consider dL and dR as lists and
        # compute the action of s2 and s3
        # from the one of l and r
        l_edges = [None]*n
        r_edges = [None]*n
        s2_edges = [None]*n
        s3_edges = [None]*n
        for o in mapping:
            i = inv_mapping[o]
            l_edges[i] = inv_mapping[dL[o]]
            r_edges[i] = inv_mapping[dR[o]]
            s2_edges[i] = inv_mapping[dS[o]]
            s3_edges[r_edges[i]] = l_edges[i]

        tcurves.append(TeichmuellerCurveOfOrigami_class(
                    mapping, inv_mapping, l_edges, r_edges, s2_edges, s3_edges))

    return tcurves

class TeichmuellerCurveOfOrigami_class(TeichmuellerCurve):
    def __init__(self, mapping, inv_mapping, l_edges, r_edges, s2_edges, s3_edges):
        N = len(s2_edges)
        ss2 = [None] * len(s2_edges)
        ss3 = [None] * len(s3_edges)
        ll = [None] * len(l_edges)
        rr = [None] * len(r_edges)
        for i in range(N):
            ss2[s2_edges[i]] = i
            ss3[s3_edges[i]] = i
            ll[l_edges[i]] = i
            rr[r_edges[i]] = i

        if mapping[0].is_orientation_cover():
            self._veech_group = EvenArithmeticSubgroup_Permutation(ss2,ss3,ll,rr)
        else:
            self._veech_group = OddArithmeticSubgroup_Permutation(ss2,ss3,ll,rr)
        self._l_edges = l_edges
        self._li_edges = ll
        self._r_edges = r_edges
        self._ri_edges = rr
        self._s2_edges = s2_edges
        self._s2i_edges = ss2
        self._s3_edges = s3_edges
        self._s3i_edges = ss3

        self._mapping = mapping
        self._inv_mapping = inv_mapping

    def __reduce__(self):
        return (TeichmuellerCurveOfOrigami_class,
                (self._mapping, self._inv_mapping,
                    self._l_edges, self._r_edges,
                    self._s2_edges, self._s3_edges))

    def _repr_(self):
        r"""
        string representation of self
        """
        return "Teichmueller curve of the origami\n%s" %self.origami()

    def stratum(self):
        r"""
        Returns the stratum of the Teichmueller curve

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2)','(1,3)')
            sage: t = o.teichmueller_curve()
            sage: t.stratum()
            H_2(2)
        """
        return self[0].stratum()

    def origami(self):
        r"""
        Returns an origmami in this teichmueller curve.

        Beware that the labels of the initial origami may have changed.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2)','(1,3)')
            sage: t = o.teichmueller_curve()
            sage: t.origami()
            (1)(2,3)
            (1,2)(3)
        """
        return self._mapping[0]

    an_element = origami

    def orbit_graph(self,
            s2_edges=True, s3_edges=True, l_edges=False, r_edges=False,
            vertex_labels=True):
        r"""
        Return the graph of action of PSL on this origami

        INPUT:

        - ``return_map`` - return the list of origamis in the orbit
        """
        G = self._veech_group.coset_graph(
                    s2_edges=s2_edges,
                    s3_edges=s3_edges,
                    l_edges=l_edges,
                    r_edges=r_edges)
        if vertex_labels:
            G.relabel(dict(enumerate(self._mapping)))
        return G

    def __getitem__(self,i):
        assert(isinstance(i,(int,Integer)))
        if i < 0 or i >= len(self._mapping):
            raise IndexError
        return self._mapping[i]

    def veech_group(self):
        r"""
        Return the veech group of the Teichmueller curve
        """
        return self._veech_group

    def plot_graph(self):
        r"""
        Plot the graph of the veech group.

        The graph corresponds to the action of the generators on the cosets
        determined by the Veech group in PSL(2,ZZ).
        """
        return self.orbit_graph().plot()

#    def plot_fundamental_domain(self):
#        from sage.geometry.hyperbolic_geometry import HH
#        from math import cos,sin,pi
#        from sage.matrix.constructor import matrix
#        from sage.all import CC
#        S=matrix([[0,-1],[1,0]])
#        T=matrix([[1,1],[0,1]])
#        g3=S*T
#        tree,gen,lifts=self._graph.spanning_tree(gq=g3)
#        pol = HH.polygon([Infinity,CC(0,1),CC(0,0),CC(-cos(pi/3),sin(pi/3))])
#        res = pol.plot(face_color='blue')
#        for m in lifts[1:]:
#            res += (m*pol).plot(face_color='green',face_alpha=0.3)
#        return res

    def cusp_representative_iterator(self):
        r"""
        Iterator over the cusp of self.

        Each term is a couple ``(o,w)`` where ``o`` is a representative of the
        cusp (an origami) and ``w`` is the width of the cusp (an integer).
        """
        cusps = []
        n = len(self._mapping)
        seen = [True]*n
        l = self._l_edges
        for i in range(n):
            if seen[i]:
                k=1
                seen[i] = False
                j = l[i]
                while j != i:
                    seen[j] = False
                    j = l[j]
                    k += 1
                yield (self._mapping[i],k)

    def cusp_representatives(self):
        return list(self.cusp_representative_iterator())

    def sum_of_lyapunov_exponents(self):
        r"""
        Returns the sum of Lyapunov exponents for this origami

        EXAMPLES::

            sage: from surface_dynamics import *

        Let us consider few examples in H(2) for which the sum is independant of
        the origami::

            sage: o = Origami('(1,2)','(1,3)')
            sage: o.stratum()
            H_2(2)
            sage: o.sum_of_lyapunov_exponents()
            4/3
            sage: o = Origami('(1,2,3)(4,5,6)','(1,5,7)(2,6)(3,4)')
            sage: o.stratum()
            H_2(2)
            sage: o.sum_of_lyapunov_exponents()
            4/3

        This is true as well for the stratum H(1,1)::

            sage: o = Origami('(1,2)','(1,3)(2,4)')
            sage: o.stratum()
            H_2(1^2)
            sage: o.sum_of_lyapunov_exponents()
            3/2
            sage: o = Origami('(1,2,3,4,5,6,7)','(2,6)(3,7)')
            sage: o.stratum()
            H_2(1^2)
            sage: o.sum_of_lyapunov_exponents()
            3/2

        ALGORITHM:

        Kontsevich-Zorich formula
        """
        K = Integer(1)/Integer(12) * sum(m*(m+2)/(m+1) for m in self.stratum().zeros())
        KK = 0
        for o in self._mapping:
            KK += sum(w/h for (h,w) in o.widths_and_heights())
        return K + Integer(1)/Integer(len(self._mapping)) * KK


    #TODO: mysterious signs and interversion problems...
    def simplicial_action_generators(self):
        r"""
        Return action of generators of the Veech group on homolgy
        """
        from sage.matrix.constructor import matrix, identity_matrix

        tree,reps,word_reps,gens = self._veech_group._spanning_tree_verrill(on_right=False)
        n = self[0].nb_squares()

        l_mat = identity_matrix(2*n)
        s_mat = matrix(2*n)
        s_mat[:n,n:] = -identity_matrix(n)

        reps = [None]*len(tree)
        reps_o = [None]*len(tree)

        reps[0] = identity_matrix(2*n)
        reps_o[0] = self.origami()

        waiting = [0]

        while waiting:
            x = waiting.pop()
            for (e0,e1,label) in tree.outgoing_edges(x):
                waiting.append(e1)
                o = reps_o[e0]
                if label == 's':
                    m = copy(s_mat)
                    m[n:,:n] = o.u().matrix().transpose()
                    reps[e1] = m * reps[e0]
                    reps_o[e1] = o.S_action()
                elif label == 'l':
                    o = o.L_action()
                    m = copy(l_mat)
                    m[:n,n:] = (~o.u()).matrix().transpose()
                    reps[e1] = m * reps[e0]
                    reps_o[e1] = o
                elif label == 'r':
                    o = o.R_action()
                    m = copy(l_mat)
                    m[n:,:n] = (~o.r()).matrix().transpose()
                    reps[e1] = m * reps[e0]
                    reps_o[e1] = o
                else:
                    raise ValueError("does know how to do only with slr")

        m_gens = []

        for (e0,e1,label) in gens:
            o = reps_o[e0]
            if label == 's':
                oo = o.S_action()
                m = copy(s_mat)
                m[n:,:n] = o.u().matrix().transpose()
            elif label == 'l':
                oo = o.L_action()
                m = copy(l_mat)
                m[:n,n:] = (~oo.u()).matrix().transpose()
            elif label == 'r':
                oo = o.R_action()
                m = copy(l_mat)
                m[:n,n:] = (~oo.r()).matrix().transpose()
            else:
                raise ValueError("does know how to do only with slr")

            ooo = reps_o[e1]
            ot, oo_renum = oo.standard_form(True)
            os, ooo_renum = ooo.standard_form(True)

            assert(ot == os)

            m_ren = (oo_renum * ~ooo_renum).matrix().transpose()
            m_s = matrix(2*n)
            m_s[:n,:n] = m_ren;
            m_s[n:,n:] = m_ren

            m_gens.append(~reps[e1] * m_s * m * reps[e0])

        return tree,reps,reps_o,gens,m_gens,word_reps




r"""
Some linear algebra routines
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

def relation_space(v):
    r"""
    Relation space of the given vector ``v``

    This is the sub vector space of `\QQ^d` given as the kernel of the map
    `n \mapsto n \cdot \lambda`. The dimension is `d - rank`.

    EXAMPLES::

        sage: from surface_dynamics.misc.linalg import relation_space
        sage: K.<sqrt2> = QuadraticField(2)
        sage: v3 = vector([sqrt2, 1, 1+sqrt2])
        sage: relation_space(v3)
        Vector space of degree 3 and dimension 1 over Rational Field
        Basis matrix:
        [ 1  1 -1]
        sage: v4 = vector([sqrt2, 1, 1+sqrt2, 1-sqrt2])
        sage: relation_space(v4)
        Vector space of degree 4 and dimension 2 over Rational Field
        Basis matrix:
        [   1    0 -1/2  1/2]
        [   0    1 -1/2 -1/2]

        sage: v = vector([1,2,5,3])
        sage: relation_space(v)
        Vector space of degree 4 and dimension 3 over Rational Field
        Basis matrix:
        [   1    0    0 -1/3]
        [   0    1    0 -2/3]
        [   0    0    1 -5/3]

    The relation space has some covariance relation with respect to matrix
    actions::

        sage: m3 = matrix(3, [1,-1,0,2,-3,4,5,-2,2])
        sage: relation_space(v3 * m3) == relation_space(v3) * ~m3.transpose()
        True
        sage: relation_space(m3 * v3) == relation_space(v3) * ~m3
        True

        sage: m4 = matrix(4, [1,-1,0,1,2,-3,0,4,5,3,-2,2,1,1,1,1])
        sage: relation_space(v4 * m4) == relation_space(v4) * ~m4.transpose()
        True
        sage: relation_space(m4 * v4) == relation_space(v4) * ~m4
        True
    """
    from sage.matrix.constructor import matrix
    try:
        m_lengths = matrix([u.vector() for u in v])
    except AttributeError:
        from sage.rings.rational_field import QQ
        v = [QQ.coerce(i) for i in v]
        m_lengths = matrix([[i] for i in v])
    return m_lengths.left_kernel()

def deformation_space(lengths):
    r"""
    Deformation space of the given ``lengths``

    This is the smallest vector space defined over `\QQ` that contains
    the vector ``lengths``. Its dimension is `rank`.

    EXAMPLES::

        sage: from surface_dynamics.misc.linalg import deformation_space
        sage: K.<sqrt2> = QuadraticField(2)
        sage: v3 = vector([sqrt2, 1, 1+sqrt2])
        sage: deformation_space(v3)
        Vector space of degree 3 and dimension 2 over Rational Field
        Basis matrix:
        [1 0 1]
        [0 1 1]
        sage: v4 = vector([sqrt2, 1, 1+sqrt2, 1-sqrt2])
        sage: deformation_space(v4)
        Vector space of degree 4 and dimension 2 over Rational Field
        Basis matrix:
        [ 1  0  1 -1]
        [ 0  1  1  1]

        sage: v = vector([1, 5, 2, 9])
        sage: deformation_space(v)
        Vector space of degree 4 and dimension 1 over Rational Field
        Basis matrix:
        [1 5 2 9]

    The deformation space has some covariance relation with respect to matrix
    actions::

        sage: m3 = matrix(3, [1,-1,0,2,-3,4,5,-2,2])
        sage: deformation_space(v3 * m3) == deformation_space(v3) * m3
        True
        sage: deformation_space(m3 * v3) == deformation_space(v3) * m3.transpose()
        True

        sage: m4 = matrix(4, [1,-1,0,1,2,-3,0,4,5,3,-2,2,1,1,1,1])
        sage: deformation_space(v4 * m4) == deformation_space(v4) * m4
        True
        sage: deformation_space(m4 * v4) == deformation_space(v4) * m4.transpose()
        True
    """
    from sage.matrix.constructor import matrix
    try:
        m_lengths = matrix([u.vector() for u in lengths])
    except AttributeError:
        from sage.rings.rational_field import QQ
        lengths = [QQ.coerce(i) for i in lengths]
        m_lengths = matrix([[i] for i in lengths])

    return m_lengths.column_space()

def deformation_cone(v):
    r"""
    Return the deformation cone of the given vector ``v``

    EXAMPLES::

        sage: from surface_dynamics.misc.linalg import deformation_cone
        sage: K.<sqrt2> = QuadraticField(2)
        sage: v3 = vector([sqrt2, 1, 1+sqrt2])
        sage: P = deformation_cone(v3)
        sage: P
        A 2-dimensional polyhedron in QQ^3 defined as the convex hull of 1 vertex and 2 rays
        sage: P.rays_list()
        [[1, 0, 1], [0, 1, 1]]
    """
    from sage.rings.all import QQ
    from sage.geometry.polyhedron.constructor import Polyhedron
    V = deformation_space(v)
    P = Polyhedron(lines=deformation_space(v).basis())
    B = Polyhedron(rays=(QQ**V.degree()).basis())
    return P.intersection(B)




# ****************************************************************************
#       Copyright (C) 2011-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

cdef gl2z_orbits(origamis, int n, int limit)
cpdef sl2z_orbits(origamis, int n, int limit)

cdef class Origami_dense_pyx:
    cdef int * _r
    cdef int * _u
    cdef int _n
    cdef object _l_edges
    cdef object _i_edges
    cdef object _name
    cdef list _pos
    cdef set _rl_frontiers
    cdef set _tb_frontiers


    cdef Origami_dense_pyx _new_c(self,int *rr_and_uu)

    cpdef _set_standard_form(self, return_map=*)

    cpdef inverse(self)
    cpdef mirror(self)
    cpdef horizontal_twist(self, width=*, cylinder=*)
    cpdef vertical_twist(self, width=*, cylinder=*)

    cdef _compute_gl2z_edges(self)

    cdef pyx_lyapunov_exponents_approx(self,
            nb_iterations, nb_experiments,
            nb_vectors, only_mean)

    cdef pyx_lyapunov_exponents_approx_with_involution(self, involution,
            nb_iterations, nb_experiments,
            nb_vectors_p, nb_vectors_m, only_mean)


cdef class PillowcaseCover_dense_pyx:
    cdef int *_g
    cdef int _n
    cdef object _l_edges
    cdef object _i_edges

    cdef PillowcaseCover_dense_pyx _new_c(self, int *g)



# coding: utf8
r"""
Permutation cover

This module deals with combinatorial data for covering of connected components of
strata of Abelian and quadratic differentials. The main feature is to be able to
compute Lyapunov exponents.
"""
# *************************************************************************
# Copyright (C) 2015-2016 Charles Fougeron <charlesfougeron@gmail.com>
#                         Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from sage.structure.sage_object import SageObject

from sage.misc.cachefunc import cached_method
from sage.libs.gap.libgap import libgap
from sage.rings.integer import Integer


class PermutationCover(SageObject):
    r"""
    A permutation with covering data

    Let `\pi` be the combinatorial data of an interval exchange transformation
    (or linear involution) on the alphabet `{1, 2, \ldots, m\}`. A cover of
    degree `d` is given by a list of permutations `\sigma_i \in S_d` for each `i
    \in \{1, 2, \ldots, m\}`.

    In order to do so, each interval on the base surface should come with an
    orientation. This orientation is automaticaly choosen by convention
    according to a clockwise orientation of the surface. The two copies of any
    interval have to be oriented alternatively in this choosen orientation and
    in the oposite to it.

    This convention is made such that the permutation associated to each interval
    is the action of the path going into the interval oriented according to
    the clockwise orientation and going out of the other one.

    This class store three attributes

    - ``_base`` -- combinatorial data of an i.e.t. or a l.i.

    - ``_degree_cover`` -- (integer) degree of the cover

    - ``_permut_cover`` -- list of permutations describing the cover

    - ``_inv_permut_cover`` -- list of inverses of ``_permut_cover``
    """
    def __init__(self, base, degree, perms):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: from surface_dynamics.interval_exchanges.cover import PermutationCover
            sage: p1 = iet.Permutation('a b c', 'c b a')
            sage: PermutationCover(p1, 2, [[0,1],[1,0],[1,0]])
            Covering of degree 2 of the permutation:
            a b c
            c b a

            sage: PermutationCover(p1, 2, [[0,1],[0,1],[0,1]])
            Traceback (most recent call last):
            ...
            ValueError: the cover is not connected

            sage: p1 = iet.Permutation('a b c d', 'b a d c')
            sage: PermutationCover(p1, 2, [[0,1],[1,0],[1,0]])
            Traceback (most recent call last):
            ...
            ValueError: the base must be irreducible

            sage: p2 = iet.Permutation('a b', 'b a')
            sage: PermutationCover(p2, 0, [[], []])
            Traceback (most recent call last):
            ...
            ValueError: the degree of the cover must be positive
        """
        if not base.is_irreducible():
            raise ValueError("the base must be irreducible")
        if degree == 0:
            raise ValueError("the degree of the cover must be positive")

        from surface_dynamics.misc.permutation import perms_are_transitive
        if not perms_are_transitive(perms):
            raise ValueError("the cover is not connected")

        from surface_dynamics.misc.permutation import perm_invert

        self._base = base.__copy__()
        self._degree_cover = degree
        self._permut_cover = perms[:]
        self._inv_permut_cover = [perm_invert(p) for p in perms]

    def __len__(self):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p1 = iet.Permutation('a b c', 'c b a')
            sage: p2 = p1.cover(['(1,2)', '(1,3)', '(2,3)'])
            sage: len(p2)
            3
        """
        return len(self._base)

    def __repr__(self):
        r"""
        A representation of the generalized permutation cover.

        INPUT:

        - ``sep`` - (default: '\n') a separator for the two intervals

        OUTPUT:

        string -- the string that represents the permutation


        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p1 = iet.Permutation('a b c', 'c b a')
            sage: p1.cover(['(1,2)', '(1,3)', '(2,3)'])
            Covering of degree 3 of the permutation:
            a b c
            c b a
        """
        s = 'Covering of degree %i of the permutation:\n'%(self._degree_cover)
        s += str(self._base)
        return s

    def __getitem__(self,i):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p1 = iet.Permutation('a b c', 'c b a')
            sage: p2 = p1.cover(['(1,2)', '(1,3)', '(2,3)'])
            sage: p2[0]
            ['a', 'b', 'c']
            sage: p2[1]
            ['c', 'b', 'a']
        """
        return self._base[i]

    def __eq__(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p1 = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p2 = iet.GeneralizedPermutation('a a b',' b c c')
            sage: p3 = iet.GeneralizedPermutation('a a b b', 'c c')
            sage: p1.cover(['(1)', '(1)', '(1)']) == p2.cover(['(1)', '(1)', '(1)'])
            True
            sage: p1.cover(['(1,2)', '', '']) == p1.cover(['(1,2)', '(1,2)', ''])
            False
            sage: p1.cover(['(1)', '(1)', '(1)']) == p3.cover(['(1)', '(1)', '(1)'])
            False
        """
        return type(self) == type(other) and \
               self._base == other._base and \
               self._degree_cover == other._degree_cover and \
               self._permut_cover == other._permut_cover

    def __ne__(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p1 = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p2 = iet.GeneralizedPermutation('a a b',' b c c')
            sage: p3 = iet.GeneralizedPermutation('a a b b', 'c c')
            sage: p1.cover(['(1)', '(1)', '(1)']) != p2.cover(['(1)', '(1)', '(1)'])
            False
            sage: p1.cover(['(1,2)', '', '']) != p1.cover(['(1,2)', '(1,2)', ''])
            True
            sage: p1.cover(['(1)', '(1)', '(1)']) != p3.cover(['(1)', '(1)', '(1)'])
            True
        """
        return type(self) != type(other) or \
               self._base != other._base or \
               self._degree_cover != other._degree_cover or \
               self._permut_cover != other._permut_cover

    def __copy__(self):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p1 = iet.Permutation('a b c', 'c b a')
            sage: p2 = p1.cover(['(1,2)', '(1,3)', '(2,3)'])
            sage: p2 == p2.__copy__()
            True
        """
        q = PermutationCover(self._base.__copy__(), self._degree_cover, self._permut_cover[:])
        return q

    def base(self):
        r"""
        Return the combinatorial data corresponding to the base of this cover

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: c = p.cover(['(1,2,3)','(1,3)','(1,2)'])
            sage: q = c.base()
            sage: q
            a b b
            c c a
            sage: q == p
            True
        """
        return self._base

    def covering_data(self, label):
        r"""
        Returns the permutation associated to the given ``label``.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = QuadraticStratum([1,1,-1,-1]).components()[0].permutation_representative()
            sage: pc = p.orientation_cover()
            sage: pc
            Covering of degree 2 of the permutation:
            0 1 2 3 3
            2 1 4 4 0

            sage: pc.covering_data(1)
            ()
            sage: pc.covering_data(3)
            (1,2)
        """
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        S = SymmetricGroup(self._degree_cover)
        return S([i+1 for i in self.covering_data_tuple(label)])

    def covering_data_tuple(self, label):
        r"""
        Returns the permutation associated to the given ``label`` as a tuple on
        `\{0, 1, \ldots, d-1\}` where `d` is the degree of the cover.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = QuadraticStratum([1,1,-1,-1]).components()[0].permutation_representative()
            sage: pc = p.orientation_cover()
            sage: pc
            Covering of degree 2 of the permutation:
            0 1 2 3 3
            2 1 4 4 0

            sage: pc.covering_data_tuple(1)
            [0, 1]
            sage: pc.covering_data_tuple(3)
            [1, 0]
        """
        return self._permut_cover[self._base.alphabet().rank(label)]

    def interval_diagram(self, sign=False):
        r"""
        Return the interval diagram.

        This is the permutation induced on the subintervals of this cover while
        we turn around the singularities. This is mainly used to compute the
        stratum of this permutation.

        OUTPUT:

        A list of lists of pairs ``(label, index of interval)`` if
        ``sign=False`` or a list of triples ``(label, index of interval,
        sign)``.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: c = p.cover(['(1,2)','(1,3)','(1,4)'])
            sage: c.interval_diagram()
            [[('a', 1), ('a', 0)],
             [('a', 2)],
             [('a', 3)],
             [('a', 0), ('b', 1), ('a', 1), ('b', 0), ('a', 2), ('b', 2)],
             [('a', 3), ('b', 3)],
             [('b', 2), ('c', 2), ('b', 0), ('c', 0), ('b', 3), ('c', 3)],
             [('b', 1), ('c', 1)],
             [('c', 3), ('c', 0)],
             [('c', 1)],
             [('c', 2)]]

            sage: c.interval_diagram(sign=True)
                [[('a', 1, -1), ('a', 0, -1)],
                 [('a', 2, -1)],
                 [('a', 3, -1)],
                 [('a', 0, 1), ('b', 1, 1), ..., ('b', 2, 1)],
                 [('a', 3, 1), ('b', 3, 1)],
                 [('b', 2, -1), ('c', 2, 1), ..., ('c', 3, 1)],
                 [('b', 1, -1), ('c', 1, 1)],
                 [('c', 3, -1), ('c', 0, -1)],
                 [('c', 1, -1)],
                 [('c', 2, -1)]]
        """
        twins, orientation = self._base._canonical_signs()
        base_diagram = self._base.interval_diagram(glue_ends=False, sign=True)
        singularities = []

        alphabet = self._base.alphabet()
        rank = alphabet.rank
        perm = lambda ss,label: self._permut_cover[rank(label)] if ss==1 else \
                                self._inv_permut_cover[rank(label)]

        for orbit in base_diagram:
            init_label = orbit[0][0]
            cover_copies = set(range(self._degree_cover))
            while cover_copies:
                d = d_init = cover_copies.pop()
                singularity = []
                while True:
                    # lift a loop from downstair
                    for base_singularity in orbit:
                        label,s = base_singularity
                        if s == -1:
                            dd = perm(s,label)[d]
                        else:
                            dd = d
                        singularity.append((label,dd,s) if sign else (label,dd))
                        d = perm(s,label)[d]

                    # if it closes, break the loop
                    if d == d_init:
                        break
                    else:
                        cover_copies.remove(d)

                singularities.append(singularity)

        return singularities

    def _delta2(self):
        r"""
        Matrix `\delta_2: C_2 -> C_1` from the `2`-cycles to the `1`-cycles.

        The matrix acts on the left. The basis of `C1` (corresponding to
        columns) is ordered first by index in the cover and then by index in the
        base.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a b b', 'c c a')
            sage: c = p.cover(['(1)', '(1)', '(1)'])
            sage: c._delta2()
            [0 0 0]
            sage: c = p.cover(['(1,2)', '(1,3)', '(1,4)'])
            sage: c._delta2()
            [ 1  1  1 -1  0  0  0 -1  0  0  0 -1]
            [-1  0  0  1  0  0  0  0  0  0  0  0]
            [ 0 -1  0  0  0  0  0  1  0  0  0  0]
            [ 0  0 -1  0  0  0  0  0  0  0  0  1]
        """
        gens = [(i,a) for i in range(self._degree_cover) for a in self._base.letters()]
        gen_indices = {g:i for i,g in enumerate(gens)}
        B = []
        p = self._base
        signs = p._canonical_signs()[1]
        for k in range(self._degree_cover):
            border = [0] * len(gens)
            for i in range(2):
                for j in range(len(p[i])):
                    if signs[i][j] == -1:
                        perm_cover = self.covering_data_tuple(p[i][j])
                        border[gen_indices[(perm_cover.index(k),p[i][j])]] += -1
                    elif signs[i][j] == 1:
                        border[gen_indices[(k,p[i][j])]] += 1
                    else:
                        RuntimeError
            B.append(border)

        from sage.matrix.constructor import matrix
        from sage.rings.integer_ring import ZZ
        return matrix(ZZ, B)

    def _delta1(self):
        r"""
        Matrix `\delta_1: C_1 -> C_0` from the `1`-chains to the `0`-chains

        The matrix acts on the left. The basis of `C_1` (corresponding to
        rows) is ordered first by index in the cover and then by index in the
        base. The basis of `C_0` is ordered as given by the method
        :meth:`interval_diagram`.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: c = p.cover(['(1)', '(1)', '(1)'])
            sage: m = c._delta1()
            sage: m
            [-1  1  0  0]
            [ 1  0 -1  0]
            [ 0  1  0 -1]
            sage: m.ncols() == len(c.profile())
            True
            sage: m.nrows() == len(c) * c._degree_cover
            True

            sage: c = p.cover(['(1,2)', '(1,3)', '(1,4)'])
            sage: (c._delta2() * c._delta1()).is_zero()
            True

            sage: a,b = QuaternionGroup().gens()
            sage: p = iet.Permutation('a b', 'b a')
            sage: c = p.cover([a, b])
            sage: assert c._delta1().nrows() == 2 * 8  # number of edges
            sage: assert c._delta1().ncols() == 4      # number of vertices
            sage: assert c._delta2().nrows() == 8      # number of faces
            sage: assert c._delta2().ncols() == 2 * 8  # number of edges
            sage: (c._delta2() * c._delta1()).is_zero()
            True

            sage: p = iet.GeneralizedPermutation('a a b c d b e', 'e f d c f')
            sage: c = p.cover(['(1,2,3,4)', '(1,3)', '(1,2)', '(2,3)', '(1,3,4)', '()'])
            sage: (c._delta2() * c._delta1()).is_zero()
            True
        """
        singularities = self.interval_diagram(sign=True)
        sing_to_index = {s:i for i,sing in enumerate(singularities) for s in sing}
        nb_sing = len(singularities)
        borders = []
        for d in range(self._degree_cover):
            for a in self._base.alphabet():
                border_side = [0] * nb_sing
                border_side[sing_to_index[(a,d,+1)]] += 1
                border_side[sing_to_index[(a,d,-1)]] += -1
                borders.append(border_side)

        from sage.matrix.constructor import matrix
        from sage.rings.integer_ring import ZZ
        return matrix(ZZ, borders)

    def profile(self):
        r"""
        Return the profile of the surface.

        The *profile* of a translation surface is the list of angles of
        singularities in the surface divided by pi.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b', 'b a')
            sage: p.cover(['(1,2)', '(1,3)']).profile()
            [6]
            sage: p.cover(['(1,2,3)','(1,4)']).profile()
            [6, 2]
            sage: p.cover(['(1,2,3)(4,5,6)','(1,4,7)(2,5)(3,6)']).profile()
            [6, 2, 2, 2, 2]

            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p.cover(['(1,2)', '()', '(1,2)']).profile()
            [2, 2, 2, 2]
        """
        from sage.combinat.partition import Partition
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup

        s = []

        base_diagram = self._base.interval_diagram(sign=True, glue_ends=True)
        p_id = SymmetricGroup(self._degree_cover).identity()
        for orbit in base_diagram:
            flat_orbit = []
            for x in orbit:
                if isinstance(x[0], tuple):
                    flat_orbit.extend(x)
                else:
                    flat_orbit.append(x)
            p = p_id
            for lab,sign in flat_orbit:
                q = self.covering_data(lab)
                if sign == -1: q = q.inverse()
                p = p*q
            for c in p.cycle_type():
               s.append(len(orbit)*c)

        return Partition(sorted(s,reverse=True))

    def is_orientable(self):
        r"""
        Test whether this permutation cover has an orientable foliation.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: from itertools import product
            sage: it = iter(product(('()', '(1,2)'), repeat=3))
            sage: next(it)
            ('()', '()', '()')

            sage: for cov in it:
            ....:     c = p.cover(cov)
            ....:     print("%28s %s" % (cov, c.is_orientable()))
            ('()', '()', '(1,2)') False
            ('()', '(1,2)', '()') False
            ('()', '(1,2)', '(1,2)') False
            ('(1,2)', '()', '()') False
            ('(1,2)', '()', '(1,2)') True
            ('(1,2)', '(1,2)', '()') False
            ('(1,2)', '(1,2)', '(1,2)') False
        """
        from surface_dynamics.interval_exchanges.template import PermutationIET
        if isinstance(self._base, PermutationIET):
            return True
        elif any(x%2 for x in self.profile()):
            return False
        else:
            # here we unfold the holonomy, i.e. try to orient each copy with +1
            # or -1
            signs = [+1] + [None] * (self._degree_cover - 1)
            todo = [0]
            A = self._base.alphabet()
            p0 = set(map(A.rank, self._base[0]))
            p1 = set(map(A.rank, self._base[1]))
            inv_letters = p0.symmetric_difference(p1)
            while todo:
                i = todo.pop()
                s = signs[i]
                assert s is not None
                for j in inv_letters:
                    ii = self._permut_cover[j][i]
                    if signs[ii] is None:
                        signs[ii] = -s
                        todo.append(ii)
                    elif signs[ii] != -s:
                        return False
            return True

    def stratum(self):
        r"""
        Stratum of the covering translation surface

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p.cover(['(1,2)', '()', '(1,2)']).stratum()
            H_1(0^4)
            sage: p.cover(['(1,2)', '(1,2)', '(1,2)']).stratum()
            Q_0(0^2, -1^4)
        """
        if self.is_orientable():
            from surface_dynamics.flat_surfaces.abelian_strata import AbelianStratum
            return AbelianStratum([(x-2)/2 for x in self.profile()])
        else:
            from surface_dynamics.flat_surfaces.quadratic_strata import QuadraticStratum
            return QuadraticStratum([x-2 for x in self.profile()])

    def genus(self):
        r"""
        Genus of the covering translation surface

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p.cover(['(1,2)', '()', '(1,2)']).genus()
            1
            sage: p.cover(['(1,2)', '(1,2)', '(1,2)']).genus()
            0

        TESTS::

            sage: from surface_dynamics import *
            sage: o = AbelianStratum([1,2,3,4]).one_component().one_origami()
            sage: assert(o.genus() == AbelianStratum([1,2,3,4]).genus())
            sage: qc = QuadraticStratum([1,2,3,4,-1,-1]).one_component()
            sage: p = qc.permutation_representative()
            sage: assert(p.orientation_cover().genus() == qc.orientation_cover_component().genus())
        """
        p = self.profile()
        return Integer((sum(p)-2*len(p))/4+1)

    @cached_method
    def _real_characters(self):
        r"""
        The real characters of the automorphism group

        OUTPUT:

        - table of characters

        - degrees

        For more information see
        :func:`~flatsurf.misc.group_representation.real_characters`.

        TESTS::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: c = p.cover(['(1,2,3)','(1,3,2)','()'])
            sage: c._real_characters()
            ([(1, 1, 1), (2, -1, -1)], [1, 1])
        """
        from surface_dynamics.misc.group_representation import real_characters
        return real_characters(self.automorphism_group())

    @cached_method
    def _cc_mats(self):
        r"""
        The projection given by the conjugacy class.

        This is cached to speed up the computation of the projection matrices.
        See :meth:`~flatsurf.misc.group_representation.conjugacy_class_matrix`
        for more informations.

        TESTS::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: c = p.cover(['(1,2,3)','(1,3,2)','()'])
            sage: c._cc_mats()
            (
            [1 0 0]  [0 1 0]  [0 0 1]
            [0 1 0]  [0 0 1]  [1 0 0]
            [0 0 1], [1 0 0], [0 1 0]
            )
        """
        from surface_dynamics.misc.group_representation import conjugacy_class_matrix
        G = self.automorphism_group()
        mats = []
        for cl in libgap(G).ConjugacyClasses():
            m = conjugacy_class_matrix(cl, self._degree_cover)
            m.set_immutable()
            mats.append(m)
        return tuple(mats)

    def isotypic_projection_matrix(self, i):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: c = p.cover(['(1,2,3)','(1,3,2)','()'])
            sage: c.isotypic_projection_matrix(0)
            [1/3   0   0 1/3   0   0 1/3   0   0]
            [  0 1/3   0   0 1/3   0   0 1/3   0]
            [  0   0 1/3   0   0 1/3   0   0 1/3]
            [1/3   0   0 1/3   0   0 1/3   0   0]
            [  0 1/3   0   0 1/3   0   0 1/3   0]
            [  0   0 1/3   0   0 1/3   0   0 1/3]
            [1/3   0   0 1/3   0   0 1/3   0   0]
            [  0 1/3   0   0 1/3   0   0 1/3   0]
            [  0   0 1/3   0   0 1/3   0   0 1/3]
        """
        from sage.matrix.special import identity_matrix
        from surface_dynamics.misc.group_representation import isotypic_projection_matrix
        m = isotypic_projection_matrix(
               self.automorphism_group(),
               self._degree_cover,
               self._real_characters()[0][i],
               self._real_characters()[1][i],
               self._cc_mats())
        return m.tensor_product(identity_matrix(len(self._base)), subdivide=False)

    def lyapunov_exponents_H_plus(self, nb_vectors=None, nb_experiments=10,
                                  nb_iterations=65536, output_file=None,
                                  return_speed=False, isotypic_decomposition=False,
                                  return_char=False, verbose=False):
        r"""
        Compute the H^+ Lyapunov exponents in  the covering locus.

        This method calls a C-library that performs renormalization (i.e. Rauzy
        induction and orthogonalization of vectors under the Kontsevich-Zorich
        cocycle). The computation might be significantly faster if
        ``nb_vectors=1`` (or if it is not provided but genus is 1) as in this
        case no orthogonalization is needed.

        INPUT:

         - ``nb_vectors`` -- the number of exponents to compute. The number of
         vectors must not exceed the dimension of the space!

         - ``nb_experiments`` -- the number of experiments to perform. It might
         be around 100 (default value) in order that the estimation of
         confidence interval is accurate enough.

         - ``nb_iterations`` -- the number of iteration of the Rauzy-Zorich
         algorithm to perform for each experiments. The default is 2^15=32768
         which is rather small but provide a good compromise between speed and
         quality of approximation.

         - ``output_file`` -- if provided (as a file object or a string) output
         the additional information in the given file rather than on the
         standard output.

         - ``return_speed`` -- wether or not return the lyapunov exponents list
         in a pair with the speed of the geodesic.

         - ``isotypic_decomposition`` -- either a boolean or a character or a list of characters.

         - ``return_char`` -- whether or not return the character corresponding to
         the isotypic component.

         - ``verbose`` -- if ``True`` provide additional informations rather than
         returning only the Lyapunov exponents (i.e. ellapsed time, confidence
         intervals, ...)


        EXAMPLES::

            sage: from surface_dynamics import *

            sage: q = QuadraticStratum([1,1,-1,-1]).one_component()
            sage: q.lyapunov_exponents_H_plus(nb_iterations=2**19) # abs tol 0.05
            [0.666666]
            sage: p = q.permutation_representative(reduced=False).orientation_cover()
            sage: c = p.lyapunov_exponents_H_plus(isotypic_decomposition=True, nb_iterations=2**19)[0]
            sage: c[0] # abs tol 0.05
            1.000000

            sage: p = iet.GeneralizedPermutation('e a a', 'b b c c d d e')
            sage: p.stratum()
            Q_0(1, -1^5)
            sage: p.alphabet()
            {'e', 'a', 'b', 'c', 'd'}
            sage: c = p.cover(['()', '(1,2)', '()', '(1,2)', '(1,2)'])
            sage: c.stratum()
            Q_1(1^2, 0^4, -1^2)
            sage: c.lyapunov_exponents_H_plus(nb_iterations=2**19) # abs tol 0.05
            [0.666666]

        Some cyclic covers (see [EKZ]_ for the formulas)::

            sage: p = iet.GeneralizedPermutation('c a a', 'b b c', alphabet='abc')
            sage: def cyclic(n,a):
            ....:     return [(i+a)%n + 1 for i in range(n)]
            sage: def cyclic_cover(n, a, b, c):
            ....:     return p.cover([cyclic(n,c), cyclic(n,a), cyclic(n, b)])

            sage: c = cyclic_cover(7,1,1,2)
            sage: c.lyapunov_exponents_H_plus(isotypic_decomposition=True, nb_iterations=2**19) # abs tol 0.05
            [[],
             [0.2857, 0.2857],
             [0.5714, 0.5714],
             [0.2857, 0.2857]]

            sage: c = cyclic_cover(7, 1, 2, 3)
            sage: lexp = c.lyapunov_exponents_H_plus(isotypic_decomposition=True, return_char=True, nb_iterations=2**19)
            sage: lexp[0]
            ([], (1, 1, 1, 1, 1, 1, 1))
            sage: lexp[1][0] # abs tol 0.05
            [0.2857, 0.2857]
            sage: lexp[1][1]
            (2, E(7) + E(7)^6, ..., E(7) + E(7)^6)
            sage: lexp[2][0] # abs tol 0.05
            [0.2857, 0.2857]
            sage: lexp[2][1]
            (2, E(7)^2 + E(7)^5, ...,  E(7)^2 + E(7)^5)
            sage: lexp[3][0] # abs tol 0.05
            [0.5714, 0.5714]
            sage: lexp[3][1]
            (2, E(7)^3 + E(7)^4, ..., E(7)^3 + E(7)^4)

        The Eierlegendewollmilchsau as a quaternionic cover of the once
        punctured torus::

            sage: p = iet.Permutation('a b', 'b a')
            sage: Q = QuaternionGroup()
            sage: a,b = Q.gens()
            sage: c = p.cover([a, b])
            sage: c.lyapunov_exponents_H_plus(nb_iterations=2**19)  # abs tol 0.05
            [1.0, 0.0, 0.0]

        REFERENCES:

        .. [EKZ] A. Eskin, M. Kontsevich, A. Zorich "Lyapunov spectrum of
                 square-tiled cyclic cover", J. Mod. Dyn. 5 (2011), no. 2, 319-353.
        """
        import surface_dynamics.interval_exchanges.lyapunov_exponents as lyapunov_exponents
        import time
        from surface_dynamics.misc.statistics import mean_and_std_dev
        from sage.matrix.special import zero_matrix

        n = len(self)

        if nb_vectors is None:
            nb_vectors = self.genus()

        if output_file is None:
            from sys import stdout
            output_file = stdout
        elif isinstance(output_file, str):
            output_file = open(output_file, "w")

        nb_vectors = int(nb_vectors)
        nb_experiments = int(nb_experiments)
        nb_iterations = int(nb_iterations)

        if nb_vectors < 0 :     raise ValueError("the number of vectors must be positive")
        if nb_vectors == 0:     return []
        if nb_experiments <= 0: raise ValueError("the number of experiments must be positive")
        if nb_iterations <= 0 : raise ValueError("the number of iterations must be positive")

        if verbose:
            output_file.write("Stratum: {}\n".format(self.stratum()))

        k = int(len(self[0]))
        gp= list(range(2*n))
        twin = list(range(2*n))
        base_twin = self._base.twin_list()
        rank = self._base.alphabet().rank
        for i in range(2):
            for j,a in enumerate(self._base[i]):
                gp[i*k+j] = rank(a)
                ii,jj = base_twin[i][j]
                twin[i*k+j] = ii*k+jj

        flatten = False
        if isotypic_decomposition:
            if isotypic_decomposition is True:
                characters = self._real_characters()[0]
            elif isinstance(isotypic_decomposition, (tuple,list)):
                if isinstance(isotypic_decomposition[0], (tuple,list)):
                    characters = map(tuple, isotypic_decomposition)
                else:
                    characters = [tuple(isotypic_decomposition)]
                    flatten = True
            else:
                raise ValueError("isotypic_decomposition must be a boolean or a list of characters")

            all_characters = self._real_characters()[0]
            for c in characters:
                if c not in all_characters:
                    raise ValueError("{} is an invalid character".format(c))

            nc = len(characters)
            na = len(all_characters)
            dim = len(self._base) * self._degree_cover
            projections = list(range(dim**2 * nc))
            dimensions = list(range(nc))
            ii = 0

            H = self._delta1().left_kernel().matrix()
            B = self._delta2()
            for i_char in range(na):
                if all_characters[i_char] not in characters:
                    continue
                M = self.isotypic_projection_matrix(i_char)
                dimension = (H*M).rank() - (B*M).rank()

                if dimension%2:
                    raise RuntimeError("get a subspace of odd dimension\n" +
                            "  i_char = {}\n".format(i_char) +
                            "  dim    = {}".format(2*dimension))
                dimensions[ii] = dimension//2
                for i in range(dim):
                    for j in range(dim):
                        projections[ii * (dim**2) + i * dim + j] = float(M[j][i])
                ii += 1
        else:
            projections = None
            dimensions = [nb_vectors]

        t0 = time.time()
        res = lyapunov_exponents.lyapunov_exponents_H_plus_cover(
            gp, k, twin, self._permut_cover, nb_experiments, nb_iterations,
            dimensions, projections, None, verbose)
        t1 = time.time()

        res_final = []

        m,d = mean_and_std_dev(res[0])
        lexp = m

        if verbose:
            from math import log, floor, sqrt
            output_file.write("sample of %d experiments\n"%nb_experiments)
            output_file.write("%d iterations (~2^%d)\n"%(
                    nb_iterations,
                    floor(log(nb_iterations) / log(2))))
            output_file.write("ellapsed time %s\n"%time.strftime("%H:%M:%S",time.gmtime(t1-t0)))
            output_file.write("Lexp Rauzy-Zorich: %f (std. dev. = %f, conf. rad. 0.01 = %f)\n"%(
                    m,d, 2.576*d/sqrt(nb_experiments)))

        if isotypic_decomposition:
            i_0 = 1
            for i_char in range(nc):
                res_int = []
                if verbose:
                    output_file.write("##### char_%d #####\n"%(i_char))
                    output_file.write("chi = {}\n".format(self._real_characters()[0][i_char]))
                    output_file.write("dim = {}\n".format(dimensions[i_char]))
                for i in range(i_0, i_0 + dimensions[i_char]):
                    m,d = mean_and_std_dev(res[i])
                    res_int.append(m)
                    if verbose:
                        output_file.write("theta_%d          : %f (std. dev. = %f, conf. rad. 0.01 = %f)\n"%(
                            i,m,d, 2.576*d/sqrt(nb_experiments)))
                res_final.append((res_int, self._real_characters()[0][i_char]) if return_char else res_int)
                i_0 += dimensions[i_char]

        else:
            for i in range(1,nb_vectors+1):
                m,d = mean_and_std_dev(res[i])
                if verbose:
                    output_file.write("theta%d           : %f (std. dev. = %f, conf. rad. 0.01 = %f)\n"%(i,m,d, 2.576*d/sqrt(nb_experiments)))
                res_final.append(m)

        res_final = res_final[0] if flatten else res_final
        if return_speed:
            return (lexp, res_final)
        else:
            return res_final

    @cached_method
    def monodromy(self):
        r"""
        Return the monodromy of this covering.

        That it to say the permutation group generated by the action of the
        fundamental group.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p.cover(['(1,2,3)', '(1,3,2)', '']).monodromy()
            Permutation Group with generators [(1,2,3), (1,3,2), ()]
            sage: p.cover(['(1,2)', '(1,3)', '']).monodromy()
            Permutation Group with generators [(1,2), (1,3), ()]
        """
        from sage.groups.perm_gps.permgroup import PermutationGroup
        return PermutationGroup([self.covering_data(a) for a in self._base.letters()], canonicalize=False)

    @cached_method
    def automorphism_group(self):
        r"""
        Return the Deck group of the cover.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
            sage: p.cover(['(1,2,3)', '(1,3,2)', '']).automorphism_group()
            Permutation Group with generators [(1,2,3), (1,3,2)]
            sage: p.cover(['(1,2)', '(1,3)', '']).automorphism_group()
            Permutation Group with generators [()]
        """
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        from sage.groups.perm_gps.permgroup import PermutationGroup

        Sd = SymmetricGroup(self._degree_cover)
        G = libgap.Subgroup(Sd, [self.covering_data(a) for a in self._base.letters()])
        C = libgap.Centralizer(Sd, G)

        return PermutationGroup(libgap.GeneratorsOfGroup(C).sage())




r"""
Python bindings for various computation of Lyapunov exponents.
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from builtins import range

from libc.stdlib cimport malloc,free
from math import isnan, isinf

cdef extern from "lyapunov_exponents.h":
    ctypedef struct quad_cover:
        pass
    ctypedef struct generalized_permutation:
        pass

    # initialisation/allocation/free
    generalized_permutation * new_generalized_permutation(int *perm, int *twin, int k, int n)
    quad_cover * new_quad_cover(generalized_permutation * gp, size_t ** sigma, size_t degree, size_t nb_vectors)
    void set_lengths(quad_cover * qcc, long double *lengths)
    void set_random_lengths_quad_cover(quad_cover * qcc)
    void set_random_vectors(quad_cover * qcc)
    void renormalize_length_quad_cover(quad_cover *qcc)

    #int check_generalized_permutation(generalized_permutation *p)
    #int check_quad_cover(quad_cover * qcc)

    void free_generalized_permutation(generalized_permutation ** gp)
    void free_quad_cover(quad_cover ** qcc)
    # print
    #void print_generalized_permutation(generalized_permutation * p)
    void print_quad_cover(quad_cover * qcc)
    void print_vectors(quad_cover * qcc)
    void print_permutation(size_t *perm, size_t degree)

    # algorithms
    #void renormalize_length_quad_cover(quad_cover * qcc)
    #void rauzy_induction_H_plus_quad_cover(quad_cover *qcc)

    int init_GS(size_t dim)
    void free_GS()
    #void orthogonalize_GS(quad_cover * qcc, double * theta)

    void lyapunov_exponents_H_plus(quad_cover *qcc, double *theta, size_t nb_induction)
    void lyapunov_exponents_isotypic(quad_cover *qcc, double *theta, size_t nb_induction, size_t nb_char, size_t *dimensions, double *proj)
    void top_lyapunov_exponents_H_plus(quad_cover *qcc, double *theta, size_t nb_iterations)

def lyapunov_exponents_H_plus_cover(
    gp, k, twin, sigma,
    nb_experiments, nb_iterations,
    dimensions, projections, lengths, verbose):
    r"""
    Compute the Lyapunov exponents of the H^+ part of the KZ-cocycle for covering locii.

    We assume that all the inputs are clean. If not, it may cause some SEGFAULT
    which would interrupt python!

    INPUT:

    - ``gp`` -- a generalized permutation given as a list of integers

    - ``k`` -- the length of the top interval

    - ``twin`` -- the twin data of the gp

    - ``sigma`` -- covering data

    - ``nb_experiments`` -- number of experimets

    - ``nb_iterations`` -- the number of iterations of the Rauzy-Zorich
      induction to perform

    - ``dimensions`` -- number of vectors we want for each component

    - ``projections`` -- isotypic projection matrices

    - ``verbose`` -- if ``True`` print additional information concerning the
      mean and standard deviation
    """
    cdef int n
    cdef int *p   # permutation
    cdef int *t   # twin
    cdef size_t **s   # sigma
    cdef size_t nc
    cdef size_t degree, i, j, nn
    cdef size_t *dim
    cdef generalized_permutation *gp_c
    cdef quad_cover *qcc
    cdef double * theta
    cdef double *proj

    if not projections and len(dimensions)>1:
        raise ValueError("dimensions without projections")
    nc = len(dimensions)

    if not isinstance(gp, (tuple, list)) or not isinstance(twin, (tuple,list)):
        raise ValueError("gp and twin should be lists")
    if len(gp)%2 or len(gp) != len(twin):
        raise ValueError("gp and twin should have the same even length")
    n = len(gp)//2
    if any(gp.count(i) != 2 for i in range(n)):
        raise ValueError("gp is not a generalized permutation")
    if set(twin) != set(range(2*n)) or \
       not all(gp[twin[i]] == gp[i] for i in range(2*n)):
        raise ValueError("invalid twin")

    if sigma is None:
        sigma = [[0]]*n
        degree = 1
    elif not isinstance(sigma, (tuple,list)):
        raise ValueError("sigma should be either a tuple or a list")
    elif len(sigma) != n:
        raise ValueError("the length of sigma (={}) should be n={}".format(len(sigma), n))
    elif not all(isinstance(l, (tuple,list)) for l in sigma):
        raise ValueError("sigma should be a list of lists")
    else:
        degree = len(sigma[0])
        sd = set(range(degree))
        if not all(len(l) == degree and set(l) == sd for l in sigma):
            raise ValueError("sigma should be a list of lists of length d that are permutations of {0, 1, ..., d}")


    p = <int *> malloc(2*n * sizeof(int))
    t = <int *> malloc(2*n * sizeof(int))

    for i in range(2*n):
        p[i] = gp[i]
        t[i] = twin[i]

    s = <size_t **> malloc(n * sizeof(size_t*))
    s[0] = <size_t *> malloc(degree * n * sizeof(size_t))
    for i in range(1,n):
        s[i] = s[0] + i * degree
    for i in range(n):
        for j in range(degree):
            s[i][j] = sigma[i][j]

    if dimensions is not None :
        nb_vectors = sum(dimensions)

    nn = int(nb_vectors) + 1 # We add one to keep track of the length of the geodesic

    theta = <double *> malloc(nn * sizeof(double))

    gp_c = new_generalized_permutation(p, t, k, n)
    qcc = <quad_cover *> new_quad_cover(gp_c, s, degree, nb_vectors)

    if lengths is None:
       set_random_lengths_quad_cover(qcc)
    else:
        l = <long double *> malloc(n * sizeof(long double))
        for i from 0 <= i < n:
            l[i] = <long double> lengths[i]
        set_lengths(qcc, l)
        free(l)

    free_generalized_permutation(&(gp_c))
    free(p)
    free(t)

    res = [[] for _ in range(nn)]

    if nb_vectors == 1:
        for i in range(nb_experiments):
            top_lyapunov_exponents_H_plus(qcc, theta, nb_iterations)
            if any(isnan(theta[j]) or isinf(theta[j]) for j in range(nn)):
                raise RuntimeError('got NaN or Inf')
            for j in range(nn):
                res[j].append(theta[j])
    else:
        init_GS(nb_vectors)

        if projections:
            dim = <size_t *> malloc(nc * sizeof(size_t))
            for i from 0 <= i < nc:
                dim[i] = int(dimensions[i])

            proj = <double *> malloc((n * degree)**2 * nc * sizeof(double))
            for i from 0 <= i < (n * degree)**2 * nc:
                 proj[i] = <double> projections[i]

            for i in range(nb_experiments):
                lyapunov_exponents_isotypic(qcc, theta, nb_iterations, nc, dim, proj)
                #cleaning some experiments which return NaN or inf as a lyapunov exponent
                if any(isnan(theta[j]) or isinf(theta[j]) for j in range(nn)):
                    raise RuntimeError('got NaN or Inf')
                for j in range(nn):
                    res[j].append(theta[j])
            free(proj)
            free(dim)

        else:
            for i in range(nb_experiments):
                lyapunov_exponents_H_plus(qcc, theta, nb_iterations)
                if any(isnan(theta[j]) or isinf(theta[j]) for j in range(nn)):
                    raise RuntimeError('got NaN or Inf')
                else:
                    for j in range(nn):
                        res[j].append(theta[j])

    free(s[0])
    free(s)

    free_quad_cover(&qcc)
    free_GS()
    free(theta)

    return res
